/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.softlayer;

import com.google.common.collect.ImmutableSet;
import com.google.inject.Module;
import org.apache.log4j.Logger;
import org.dasein.cloud.AbstractCloud;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.jclouds.softlayer.compute.SoftLayerComputeServices;
import org.jclouds.compute.ComputeServiceContext;
import org.jclouds.compute.ComputeServiceContextFactory;
import org.jclouds.compute.domain.OperatingSystem;
import org.jclouds.concurrent.config.ExecutorServiceModule;
import org.jclouds.logging.log4j.config.Log4JLoggingModule;
import org.jclouds.rest.RestContext;
import org.jclouds.softlayer.SoftLayerAsyncClient;
import org.jclouds.softlayer.SoftLayerClient;
import org.jclouds.softlayer.domain.ProductPackage;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import static org.jclouds.concurrent.MoreExecutors.sameThreadExecutor;

public class SoftLayer extends AbstractCloud {
    static private final Logger logger = Logger.getLogger(SoftLayer.class);

    private static final String SOFTLAYER_VIRTUALGUEST_PACKAGE_NAME = "Cloud Server";

    public SoftLayer() { }
    
    @Override
    public @Nonnull String getCloudName() {
        return "SoftLayer";
    }

    public @Nonnull RestContext<SoftLayerClient,SoftLayerAsyncClient> getCloudClient() throws CloudException {
        ProviderContext ctx = getContext();
        
        if( ctx == null ) {
            throw new CloudException("No context has been defined for this request");
        }
        String user = ctx.getAccountNumber();
        String key = new String(ctx.getAccessPrivate());

        ComputeServiceContext softlayer = (new ComputeServiceContextFactory()).createContext("softlayer", user, key, ImmutableSet.<Module>of(new Log4JLoggingModule(), new ExecutorServiceModule(sameThreadExecutor(), sameThreadExecutor())));
        return softlayer.getProviderSpecificContext();
    }
    
    @Override
    @SuppressWarnings("NullableProblems")
    public @Nonnull SoftLayerComputeServices getComputeServices() {
        return new SoftLayerComputeServices(this);
    }
    
    @Override
    public @Nonnull SoftLayerDC getDataCenterServices() {
        return new SoftLayerDC(this);
    }

    public @Nonnull String getVirtualGuestPackageName() {
        return SOFTLAYER_VIRTUALGUEST_PACKAGE_NAME;
    }

    public @Nonnull String getDomain() {
        return "dasein.net";
    }

    /**
     * ex: for private (ex. don't share hardware) use "Private [0-9]+ x ([.0-9]+) GHz Core[s]?"
     * ex: for private and public use ".*[0-9]+ x ([.0-9]+) GHz Core[s]?"
     * @return regex for CPU parsing
     */
    public @Nonnull String getCpuRegex() {
        return "[0-9]+ x ([0-9.]+) GHz Core[s]?";
    }

    public @Nonnegative float getPortSpeed() {
        // 10Mbit is free, 100 has recurringFee=10, 1000 has recurringFee=20
        return 10;
    }

    public @Nonnull String getDisk0Type() {
        return "LOCAL";
    }

    public @Nonnegative long getLoginTimeoutMs() {
        return 60*60*1000L; // Softlayer can be slow.
    }

    @Override
    public @Nonnull String getProviderName() {
        return "SoftLayer";
    }
    
    public @Nonnull Architecture toArchitecture(OperatingSystem os) {
        return os.is64Bit() ? Architecture.I64 : Architecture.I32;
    }
    
    public @Nonnull Platform toPlatform(OperatingSystem os) {
        if( os == null ) {
            return Platform.UNKNOWN;
        }

        switch (os.getFamily()) {
           case UBUNTU: return Platform.UBUNTU;
           case DEBIAN: return Platform.DEBIAN;
           case SOLARIS:return Platform.SOLARIS;
           case FEDORA: return Platform.FEDORA_CORE;
           case RHEL: return Platform.RHEL;
           case FREEBSD: return Platform.FREE_BSD;
           case OPENBSD: return Platform.OPEN_BSD;
           case CENTOS: return Platform.FEDORA_CORE;
           case WINDOWS:return Platform.WINDOWS;
           default: return Platform.guess(os.getName());
        }
    }

    private ProductPackage productPackage;

    /**
     * Get the product package.
     * @return The product package to use for launching virtual guests.
     * @throws CloudException if the ProductPackage cannot be retrieved
     */
    public synchronized @Nonnull ProductPackage getProductPackage() throws CloudException {

        if ( productPackage!=null ) {
            return productPackage;
        }

        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = getCloudClient();
        try {
            for( ProductPackage pp: ctx.getApi().getAccountClient().getActivePackages() ) {
               if( getVirtualGuestPackageName().equals(pp.getName()) ) {
                   productPackage = ctx.getApi().getProductPackageClient().getProductPackage(pp.getId());
                   break;
               }
            }
        } catch( Throwable t ) {
            final String message = "Failed to obtain SoftLayer product package:"+getVirtualGuestPackageName()+", "+t.getMessage();
            logger.warn(message);
            if( logger.isDebugEnabled() ) {
                t.printStackTrace();
            }
            throw new CloudException(message);
        } finally {
            ctx.close();
        }

       return productPackage;
   }
    
    @Override
    public @Nullable String testContext() {
        try {
            ProviderContext ctx = getContext();

            if( ctx == null ) {
                return null;
            }
            if( !getComputeServices().getVirtualMachineSupport().isSubscribed() ) {
                return null;
            }
            return ctx.getAccountNumber();
        }
        catch( Throwable t ) {
            logger.warn("Failed to test SoftLayer connection context: " + t.getMessage());
            if( logger.isDebugEnabled() ) {
                t.printStackTrace();
            }
            return null;
        }
    }
}
