/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.softlayer;

import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.dc.DataCenter;
import org.dasein.cloud.dc.DataCenterServices;
import org.dasein.cloud.dc.Region;
import org.jclouds.softlayer.domain.Datacenter;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;

public class SoftLayerDC implements DataCenterServices {
   private SoftLayer provider = null;

    SoftLayerDC(@Nonnull SoftLayer provider) {
        this.provider = provider;
    }

    @Override
    public @Nonnull String getProviderTermForDataCenter(@Nonnull Locale locale) {
        return "datacenter";
    }

    @Override
    public @Nonnull String getProviderTermForRegion(@Nonnull Locale locale) {
        return "region";
    }

    @Override
    public @Nullable DataCenter getDataCenter(@Nonnull String dataCenterId) throws InternalException, CloudException {
        populateRegionMap();
        for( Collection<DataCenter> dcs : regions.values() ) {
            for( DataCenter dc : dcs ) {
                if( dataCenterId.equals(dc.getProviderDataCenterId()) ) {
                    return dc;
                }
            }
        }
        return null;
    }

    @Override
    public @Nullable Region getRegion(@Nonnull String regionId) throws InternalException, CloudException {
        populateRegionMap();
        for( Region region : listRegions() ) {
            if( region.getProviderRegionId().equals(regionId) ) {
                return region;
            }
        }
        return null;
    }

    @Override
    public @Nonnull Collection<DataCenter> listDataCenters(@Nonnull String regionId) throws InternalException, CloudException {
        populateRegionMap();
        Region region = getRegion(regionId);

        if( region == null ) {
            throw new CloudException("No such region: " + regionId);
        }
        Collection<DataCenter> dataCenters = regions.get(region);
        
        if ( dataCenters == null ) {
           return Collections.emptyList();
        }
        return Collections.unmodifiableCollection(dataCenters);
    }

    @Override
    public @Nonnull Collection<Region> listRegions() throws InternalException, CloudException {
        populateRegionMap();
        return Collections.unmodifiableSet(regions.keySet());
    }

    static private Map<Region,Collection<DataCenter>> regions = null;

    private synchronized void populateRegionMap() throws CloudException {
       if( regions == null ) {
            Map<Region,Collection<DataCenter>> regionMap = new HashMap<Region,Collection<DataCenter>>();
            for( Datacenter slDataCenter : provider.getProductPackage().getDatacenters() ) {
                for( org.jclouds.softlayer.domain.Region slRegion : slDataCenter.getRegions() ) {
                   if( slRegion.getKeyname().equals("FIRST_AVAILABLE") ) continue;

                   Region region = new Region();
                   region.setActive(true);
                   region.setAvailable(true);
                   region.setProviderRegionId(slRegion.getKeyname());
                   region.setName(slRegion.getDescription());
                   if( region.getName() == null ) {
                      region.setName(region.getProviderRegionId());
                   }

                   Collection<DataCenter> datacenters = regionMap.get(region);
                   if ( datacenters == null ) {
                      datacenters = new HashSet<DataCenter>();
                   }

                   DataCenter dc = new DataCenter();
                   dc.setActive(true);
                   dc.setAvailable(true);
                   dc.setName(slDataCenter.getName());
                   dc.setProviderDataCenterId(Integer.toString(slDataCenter.getId()));
                   dc.setRegionId(region.getProviderRegionId());

                   datacenters.add(dc);
                   regionMap.put(region,datacenters);
                }

            }
            regions = regionMap;
      }
   }
}
