/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.softlayer.compute;

import com.google.common.base.Function;
import com.google.common.collect.Sets;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.MachineImageFormat;
import org.dasein.cloud.compute.MachineImageState;
import org.dasein.cloud.compute.MachineImageSupport;
import org.dasein.cloud.compute.MachineImageType;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.softlayer.SoftLayer;
import org.jclouds.compute.domain.Image;
import org.jclouds.rest.AuthorizationException;
import org.jclouds.rest.RestContext;
import org.jclouds.softlayer.SoftLayerAsyncClient;
import org.jclouds.softlayer.SoftLayerClient;
import org.jclouds.softlayer.compute.functions.ProductItemToImage;
import org.jclouds.softlayer.domain.ProductItem;
import org.jclouds.softlayer.predicates.ProductItemPredicates;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;
import java.util.Set;

import static com.google.common.collect.Iterables.filter;
import static com.google.common.collect.Iterables.transform;

public class SoftLayerImage implements MachineImageSupport {

    private SoftLayer provider;

    SoftLayerImage(SoftLayer softlayer) {
        provider = softlayer;
    }

    @Override
    public void remove(@Nonnull String imageId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("SoftLayer does not allow you to delete images.");
    }

    @Override
    public @Nonnull Collection<MachineImage> searchMachineImages(@Nullable String keyword, @Nullable Platform platform, @Nullable Architecture architecture) throws InternalException, CloudException {
        ArrayList<MachineImage> result = new ArrayList<MachineImage>();
        
        for( MachineImage image :  listPublicImages() ) {
            if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
                continue;
            }
            if( platform != null && !platform.equals(Platform.UNKNOWN) ) {
                Platform mine = image.getPlatform();

                if( platform.isWindows() && !mine.isWindows() ) {
                    continue;
                }
                if( platform.isUnix() && !mine.isUnix() ) {
                    continue;
                }
                if( platform.isBsd() && !mine.isBsd() ) {
                    continue;
                }
                if( platform.isLinux() && !mine.isLinux() ) {
                    continue;
                }
                if( platform.equals(Platform.UNIX) ) {
                    if( !mine.isUnix() ) {
                        continue;
                    }
                }
                else if( !platform.equals(mine) ) {
                    continue;
                }
            }
            if( keyword != null && !keyword.equals("") ) {
                keyword = keyword.toLowerCase();
                if( !image.getProviderMachineImageId().toLowerCase().contains(keyword) ) {
                    if( !image.getName().toLowerCase().contains(keyword) ) {
                        if( !image.getDescription().toLowerCase().contains(keyword) ) {
                            continue;
                        }
                    }
                }
            }
            result.add(image);
        }
        return result;
    }


    @Override
    public @Nullable MachineImage getMachineImage(@Nonnull String imageId) throws InternalException, CloudException {
        for( MachineImage image : listPublicImages() ) {
            if(imageId.equals(image.getProviderMachineImageId()) ) {
                return image;
            }
        }
        return null;
    }

    /**
     * An 'image' is a product item with a category code 'os'
     */
    @Override
    public @Nonnull String getProviderTermForImage(@Nonnull Locale locale) {
        return "product item";
    }

    @Override
    public @Nonnull Collection<String> listShares(@Nonnull String imageId) throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public @Nonnull AsynchronousTask<String> imageVirtualMachine(@Nonnull String serverId, @Nonnull String name, @Nonnull String description) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    @Override
    public @Nonnull Collection<MachineImage> listMachineImages() throws InternalException, CloudException {
        return Collections.emptyList();
    }
    
    private @Nonnull Collection<MachineImage> listPublicImages() throws InternalException, CloudException {
        Set<ProductItem> items = provider.getProductPackage().getItems();
        
        if( items == null ) {
            return Collections.emptySet();
        }

        Iterable<MachineImage> images = transform(filter(items, ProductItemPredicates.categoryCode("os")), toImage());
        return Sets.newLinkedHashSet(images);
    }

    private @Nonnull Function<ProductItem,MachineImage> toImage() {
        return new Function<ProductItem, MachineImage>() {
            public @Nullable MachineImage apply(@Nullable ProductItem item) {
                if( item == null ) {
                    return null;
                }
                ProviderContext ctx = provider.getContext();
                
                if( ctx == null ) {
                    throw new RuntimeException("Unable to load machine image without a context");
                }
                String regionId = ctx.getRegionId();
                
                if( regionId == null ) {
                    return null;
                }
                Image image = new ProductItemToImage().apply(item);

                MachineImage result = new MachineImage();

                String providerId = image.getId();
                
                if( providerId == null ) {
                    return null;
                }
                result.setProviderMachineImageId(providerId);
                
                String name = image.getName();
                
                if( name == null ) {
                    name = image.getDescription();
                    if( name == null ) {
                        name = providerId;
                    }
                }
                result.setName(name);
                
                String description = image.getDescription();

                if( description == null ) {
                    description = name;
                }
                result.setDescription(description);
                result.setCurrentState(MachineImageState.ACTIVE);
                result.setProviderRegionId(regionId);
                result.setArchitecture(provider.toArchitecture(image.getOperatingSystem()));
                result.setPlatform(provider.toPlatform(image.getOperatingSystem()));
                result.setType(MachineImageType.STORAGE);
                result.setProviderOwnerId("--public--");
                result.setSoftware("");

                return result;
            }
        };
    }

    @Override
    public void downloadImage(@Nonnull String machineImageId, @Nonnull OutputStream toOutput) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean hasPublicLibrary() {
        return true;
    }

    @Override
    public @Nonnull AsynchronousTask<String> imageVirtualMachineToStorage(String vmId, String name, String description, String directory) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public @Nonnull String installImageFromUpload(@Nonnull MachineImageFormat format, @Nonnull InputStream imageStream) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean isImageSharedWithPublic(@Nonnull String machineImageId) throws CloudException, InternalException {
        return true;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();
        try {
            provider.getProductPackage();
            return true;
        }
        catch( AuthorizationException e ) {
            return false;
        }
        catch( RuntimeException e ) {
            return false;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public @Nonnull Iterable<MachineImage> listMachineImagesOwnedBy(@Nullable String accountId) throws CloudException, InternalException {
        if( !"--public--".equals(accountId) ) {
            return listPublicImages();
        }
        return Collections.emptyList();
    }

    @Override
    public @Nonnull Iterable<MachineImageFormat> listSupportedFormats() throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }

    @Override
    public @Nonnull String registerMachineImage(@Nonnull String atStorageLocation) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public void shareMachineImage(@Nonnull String machineImageId, @Nonnull String withAccountId, boolean allow) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean supportsCustomImages() {
        return false;
    }

    @Override
    public boolean supportsImageSharing() {
        return false;
    }

    @Override
    public boolean supportsImageSharingWithPublic() {
        return false;
    }

    @Override
    public @Nonnull String transfer(@Nonnull CloudProvider fromCloud, @Nonnull String machineImageId) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No supported");
    }
}
